classdef Channel < matlab.mixin.SetGet
% Channel Class that implements an XCP protocol user channel.

% Copyright 2012-2021 The MathWorks, Inc.

properties (SetAccess = 'private')
    ServerName            % The name of the XCP server node to which this channel connects.
    A2LFileName           % The name of the A2L file that describes the XCP target node.
    TransportLayer        % The type of transport layer used for the XCP connection.
    TransportLayerDevice  % Structure containing details on the XCP transport layer connection.
end

properties
    SeedKeyDLL            % A DLL file containing the seed and key access algorithm.
end

properties (SetAccess = 'private', GetAccess = 'private')
    A2LFileObject        % An A2L object.
    ProtocolObject       % An xcp.Protocol object that implements the XCP protocol messaging.
    ServerInfo           % A structure of server information acquired from the server.
    ConnectState         % Stores the current server connection state of the channel.
    MeasurementState     % Stores the current state of measurement activities on the channel.
    DAQListObject        % An array of all currently configured DAQ list objects.
    DAQData              % A structure with array values of stored DAQ data for measurements.
    DAQTimer             % A timer object that runs the periodic DAQ tasks.
    DAQTimerListener     % Listener for the DAQ timer events.
    DAQTimerRate         % The rate at which to run the DAQ periodic task.
    STIMListObject       % An array of all currently configured STIM list objects.
    STIMTimer            % A timer object that runs the periodic STIM tasks.
    STIMTimerListener    % Listener for the STIM timer events.
    STIMTimerRate        % The rate at which to run the STIM periodic task.
    SeedKeyDLLFullPath   % Full path to the seed and key access DLL file.
    api_xcp_daqObject    % MEX Interface Class for Receiving data for Measurements
    DAQClockResp
    GetResolutionResp
 %    DataAvailCallback
end

properties (Hidden, SetAccess = 'private')
    SlaveName             % Original name of the server node(ServerName). Keep it as a hidden property to maintain release compatibility.
end

properties (Dependent, SetAccess = 'private', GetAccess = 'private')
    NumberOfDAQLists         % The number of configured DAQ lists.
    NumberOfSTIMLists        % The number of configured STIM lists.
    NumberOfMeasurementLists % The number of configured DAQ and STIM lists.
end


methods
    
    function obj = Channel(a2l, transportLayer, varargin)
    % Channel Creates and returns an XCP Channel object.
    
        % Convert string inputs to character vectors.
        [transportLayer, varargin{:}] = convertStringsToChars(transportLayer, varargin{:});
        
        % Validate the A2L argument.
        validateattributes(a2l, {'xcp.A2L'}, {'scalar', 'nonempty'}, 'xcp.Channel', 'A2L file object', 1);
        
        % Convert the transport layer argument to all upper case.
        transportLayer = upper(transportLayer);
        % Validate the transport layer argument. 
        transportLayer = validatestring(transportLayer, {'CAN', 'CAN FD', 'TCP', 'UDP'}, 'xcp.Channel', 'transport layer', 2);
        
        % Set the base property values from the input arguments.
        obj.A2LFileObject = a2l;
        obj.A2LFileName = a2l.FileName;
        obj.ServerName = a2l.ServerName;
        obj.TransportLayer = transportLayer;
        
        obj.SlaveName = obj.ServerName;
        obj.api_xcp_daqObject  = clib.libapi_xcp_daq.api_xcp_daq;
        
        switch transportLayer
            case {'CAN', 'CAN FD'}
                % Perform an argument count check.
                narginchk(4, 5);
                
                % Validate the variable arguments.
                validateattributes(varargin{1}, {'char'}, {'nonempty'}, 'xcp.Channel', 'transport layer device vendor', 3);
                validateattributes(varargin{2}, {'char'}, {'nonempty'}, 'xcp.Channel', 'transport layer device', 4);
                if nargin == 5
                    validateattributes(varargin{3}, {'numeric'}, {'nonempty', 'integer', 'real', 'finite', 'nonnan', 'positive'}, ...
                        'xcp.Channel', 'transport layer device channel index', 5);
                end
                
                % CAN/CAN FD transport layer: get hardware adaptor details.
                obj.TransportLayerDevice.Vendor = varargin{1};
                obj.TransportLayerDevice.Device = varargin{2};
                if nargin == 5
                    obj.TransportLayerDevice.ChannelIndex = varargin{3};
                else
                    obj.TransportLayerDevice.ChannelIndex = [];
                end
                
                % If the transport layer is CAN FD.
                if obj.TransportLayer == "CAN FD"
                    % Throw an error if the device vendor is PEAK-System.
                    if obj.TransportLayerDevice.Vendor == "PEAK-System"
                        error(message('vnt:XCP:CANFDDoesNotSupportPEAK'));
                    end
                    % Throw an error if CAN FD info from A2L file is empty.
                    if isempty(obj.A2LFileObject.TransportLayerCANInfo.CANFD)
                        error(message('vnt:XCP:NoCANFDInfo'));
                    end
                end

            case 'TCP'
                % Perform an argument count check.
                narginchk(2, 4);
                
                % Process if the connection information was provided as
                % input or is coming from the A2L file.
                switch nargin
                    case 2
                        if ~isempty(a2l.TransportLayerTCPInfo)
                            obj.TransportLayerDevice.AddressString = a2l.TransportLayerTCPInfo.AddressString;
                            obj.TransportLayerDevice.Port = uint16(a2l.TransportLayerTCPInfo.Port);
                        else
                            error(message('vnt:XCP:MissingHostPortInfo'));
                        end
                    case 4
                        obj.TransportLayerDevice.AddressString = varargin{1};
                        obj.TransportLayerDevice.Port = uint16(varargin{2});
                end

            case 'UDP'
                % Perform an argument count check.
                narginchk(2, 4);

                % Process if the connection information was provided as
                % input or is coming from the A2L file.
                switch nargin
                    case 2
                        if ~isempty(a2l.TransportLayerUDPInfo)
                            obj.TransportLayerDevice.AddressString = a2l.TransportLayerUDPInfo.AddressString;
                            obj.TransportLayerDevice.Port = uint16(a2l.TransportLayerUDPInfo.Port);
                        else
                            error(message('vnt:XCP:MissingHostPortInfo'));
                        end
                    case 4
                        obj.TransportLayerDevice.AddressString = varargin{1};
                        obj.TransportLayerDevice.Port = uint16(varargin{2});
                end
        end
        
        % Initialize internal states.
        obj.ConnectState = false;
        obj.MeasurementState = false;
        
        % Initialize the measurement list properties.
        obj.DAQListObject = xcp.MeasurementList.empty();
        obj.DAQData = containers.Map;
        obj.DAQTimerRate = 0.005;
        obj.STIMListObject = xcp.MeasurementList.empty();
        obj.STIMTimerRate = 0.010;
    end
     
    function delete(obj)
    % delete Clean up and remove the object.
    %
    %   delete(OBJ) deletes the XCP channel object OBJ. Any activity is
    %   stopped and connection to a server module is disconnected.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       delete(channelObj);
    %
    %   See also VNT.
        
        % Disconnect from the server if the connection is active.
        if obj.isConnected()
            obj.disconnect();
        end
    end
    
    
    function connect(obj)
    % connect Starts an active connection to the server module.
    %
    %   connect(OBJ) actively connects the XCP channel object OBJ to the server
    %   module that was specified during creation of the object. The
    %   required XCP commands necessary to establish a connection are sent
    %   to the server and any applicable responses are received and
    %   processed.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %
    %   See also VNT.
    
        % Perform an argument count check.
        narginchk(1, 1);
        
        % Validate current state conditions.
        obj.errorWhenConnected();

        % Build a protocol object for communicating with the server.
        try
            obj.ProtocolObject = xcp.Protocol( ...
                obj.A2LFileObject, ...
                obj.TransportLayer, ...
                obj.TransportLayerDevice);
        catch err
            error(message('vnt:XCP:UnableToCreateProtocolObject'));
        end
        
        % Connect to the server.
        [~, resource, commModeBasic, maxCTO, maxDTO] = obj.ProtocolObject.connect(xcp.Utility.USER_DEFINED);
        % Store the returned information in this channel.
        obj.ServerInfo.Resource = resource;
        obj.ServerInfo.CommModeBasic = commModeBasic;
        % CTO (Command Transfer Object) is used for transferring generic control commands.
        % It is used for protocol commands (client to server), command responses (server to client),
        % error packets (server to client), event packets (server to client), and service request packets (server to client).
        obj.ServerInfo.MaxCTO = uint32(maxCTO);
        % DTO (Data Transfer Object) is used for transmitting synchronous data acquisition data (server to client)
        % and synchronous data stimulation data (client to server).
        obj.ServerInfo.MaxDTO = uint32(maxDTO);
        
        % Query for current server status.
        [~, currentSessionStatus, currentResourceProtectionStatus, sessionConfigurationID] = obj.ProtocolObject.getStatus();
        % Store the returned information in this channel.
        obj.ServerInfo.CurrentSessionStatus = currentSessionStatus;
        obj.ServerInfo.CurrentResourceProtectionStatus = currentResourceProtectionStatus;
        obj.ServerInfo.SessionConfigurationID = sessionConfigurationID;
        
        % If DAQ is available and locked, unlock it.
        if obj.ServerInfo.Resource.DAQ && obj.ServerInfo.CurrentResourceProtectionStatus.DAQ
            obj.unlockResource('DAQ');
        end
        
        % If STIM is available and locked, unlock it.
        if obj.ServerInfo.Resource.STIM && obj.ServerInfo.CurrentResourceProtectionStatus.STIM
            obj.unlockResource('STIM');
        end
        % If CAL is available and locked, unlock it.
        if obj.ServerInfo.Resource.CAL && obj.ServerInfo.CurrentResourceProtectionStatus.CAL
            obj.unlockResource('CAL');
        end
        
        % Modify the internal state.
        obj.ConnectState = true;
       [getClockResp] = obj.ProtocolObject.getDAQClock(); 
       obj.DAQClockResp = getClockResp;
       [getResolutionResp]= obj.ProtocolObject.getDAQResolutionInfo();
       obj.GetResolutionResp= getResolutionResp;
%     obj.configureCallback(obj.ProtocolObject.TransportLayerObject,"datagram",1, obj.DataAvailCallback)
    end
    
    function disconnect(obj)
    % disconnect Stops an active connection to the server module.
    %
    %   disconnect(OBJ) disconnects the XCP channel object OBJ from the server
    %   module that was specified during creation of the object. The
    %   required XCP commands necessary to close a connection are sent
    %   to the server and any applicable responses are received and
    %   processed.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %       disconnect(channelObj);
    %
    %   See also VNT.
    
        % Perform an argument count check.
        narginchk(1, 1);
        
        % Validate current state conditions.
        obj.errorWhenDisconnected();

        % If measurements are active, stop them.
        if obj.isMeasurementRunning()
            obj.stopMeasurement();
        end

        % If the channel is connected, disconnect it.
        if obj.isConnected()
            obj.ProtocolObject.disconnect();
            obj.ProtocolObject = [];
        end
        
        % Modify the internal state.
        obj.ConnectState = false;
    end
    
    
    function createMeasurementList(obj, resource, eventName, measurementNames)
    % createMeasurementList Configure a DAQ or STIM list on the XCP channel.
    %
    %   createMeasurementList(OBJ, RESOURCE, EVENT_NAME, MEASUREMENT_NAMES)
    %   configures a dynamic DAQ or STIM list on the XCP channel object
    %   OBJ. The RESOURCE must be specified as a string to denote this a
    %   'DAQ' or 'STIM' list. The EVENT_NAME is also a string that is the
    %   name of the event on which to configure this list. Measurements are
    %   specified as MEASUREMENT_NAMES as a string or cell array of strings
    %   to define which measurements are configured for this list.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %       createMeasurementList(channelObj, 'DAQ', 'Event1', 'Measurement1');
    %       createMeasurementList(channelObj, 'STIM', 'Event2', {'Measurement1', 'Measurement1'});
    %
    %   See also VNT.
    
        % Perform an argument count check.
        narginchk(4, 4);

        % Convert string inputs to character vectors.
        [resource, eventName, measurementNames] = convertStringsToChars(resource, eventName, measurementNames);

        % Validate current state conditions.
        obj.errorWhenDisconnected();
        obj.errorWhenMeasurementActive();
        
        % Force the direction argument to be all caps because that is how
        % we want the values represented inside the channel.
        resource = upper(resource);
        % Validate the resource argument.
        resource = validatestring(resource, {'DAQ', 'STIM'}, 'xcp.createMeasurementList', 'direction', 2);

        % Validate the event argument.
        eventName = validatestring(eventName,  obj.A2LFileObject.Events, 'xcp.createMeasurementList', 'event name', 3);

        % Create a cell array out of the measurement names argument. This
        % argument may be sent as a single string, but we always reference
        % this argument as a cell array.
        if ~iscell(measurementNames)
            measurementNames = {measurementNames};
        end
        
        % Validate each entry in the measurements argument.
        for ii = 1:numel(measurementNames)
            measurementNames{ii} = validatestring(measurementNames{ii}, obj.A2LFileObject.Measurements, ...
                'xcp.createMeasurementList', 'measurement names', 4);
        end
        
        % Verify that the requested resource is an available resource on
        % this server. For example, error if STIM was requested but STIM is
        % not supported by the server.
        if ~obj.ServerInfo.Resource.(resource)
            error(message('vnt:XCP:ResourceNotAvailableByServer', resource));
        end
        
        % Verify that the specified event is not already in use for the
        % given resource. We only allow an event to be used once for DAQ
        % and once for STIM at a time.
        for ii = 1:obj.(['NumberOf' resource 'Lists'])
            % Check through each list already configured for the given
            % resource type to see if the list is using the requested event.
            if strcmpi(eventName, obj.([resource 'ListObject'])(ii).EventInfo.Name)
                error(message('vnt:XCP:EventAlreadyUsedForThisResource', eventName, resource));
            end
        end
        
        % Verify that the specified measurements are not already in use for
        % the given resource. We only allow a measurement to be used once
        % for DAQ and once for STIM at a time.
        for ii = 1:numel(measurementNames)
            % Check each resource list.
            for jj = 1:obj.(['NumberOf' resource 'Lists'])
                % Use logical indexing to check for the measurement.
                if any(strcmpi(measurementNames{ii}, obj.([resource 'ListObject'])(jj).MeasurementNames))
                    error(message('vnt:XCP:MeasurementAlreadyUsedForThisResource', measurementNames{ii}, resource));
                end
            end
        end
        
        % Get the detailed event information from the A2L file.
        eventInfo = obj.A2LFileObject.getEventInfo(eventName);
        
        % Validate that this event supports the specified resource. Some
        % events can be used for both STIM and DAQ while others only
        % support one option.
        if ~strcmpi(resource, eventInfo.Direction) && ~strcmpi(eventInfo.Direction, 'DAQ_STIM')
            error(message('vnt:XCP:EventDirectionNotSupported', eventInfo.Name, resource));
        end
        
        % If this is a STIM list, verify that the specified event is a
        % periodic/time-based event. Only time-based events are supported for
        % STIM in the current implementation, but the A2L file may have
        % non-periodic events configured as possible for STIM.
        if strcmpi(resource, 'STIM') && eventInfo.ChannelTimeCycleInSeconds == 0
            error(message('vnt:XCP:EventNotPeriodic', eventInfo.Name, resource));
        end
        
        % Get the detailed measurement information from the A2L file for
        % each specified measurement.
        measurementInfo = [];
        for ii = 1:numel(measurementNames)
            measurementInfo = [measurementInfo obj.A2LFileObject.getMeasurementInfo(measurementNames{ii})]; %#ok<AGROW>
        end
        
        % Verify that each of the specified measurements are smaller in
        % size than the maximum allowed by the server.
        for ii = 1:numel(measurementNames)
            if measurementInfo(ii).sizeInBytes() > obj.A2LFileObject.DAQInfo.MaxODTEntrySizeDAQ
                error(message('vnt:XCP:MeasurementExceedsMaximumODTEntrySize', ...
                    measurementNames{ii}, ...
                    obj.A2LFileObject.DAQInfo.MaxODTEntrySizeDAQ));
            end
        end
        
        % Create a new measurement list for the specified resource.
        obj.([resource 'ListObject']) = [obj.([resource 'ListObject']) ...
            xcp.MeasurementList(obj.A2LFileObject, resource, eventInfo, measurementInfo)];
    end
    
    function freeMeasurementLists(obj)
    % freeMeasurementLists Remove all DAQ and STIM lists from the XCP channel.
    %
    %   freeMeasurementLists(OBJ) deletes all configured DAQ and STIM lists
    %   configured on the XCP channel object OBJ.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %       createMeasurementList(channelObj, 'DAQ', 'Event1', 'Measurement1');
    %       freeMeasurementLists(channelObj);
    %
    %   See also VNT.
    
        % Perform an argument count check.
        narginchk(1, 1);

        % Validate current state conditions.
        obj.errorWhenMeasurementActive();
        obj.errorWhenNoMeasurementListsConfigured();
        
        % Erase the stored measurement lists.
        obj.DAQListObject = xcp.MeasurementList.empty();
        obj.STIMListObject = xcp.MeasurementList.empty();
    end
    
    function viewMeasurementLists(obj)
    % viewMeasurementLists View the configured DAQ and STIM lists.
    %
    %   viewMeasurementLists(OBJ) displays information to the command line
    %   about all of the currently configured DAQ and STIM lists on the XCP
    %   channel object OBJ.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %       createMeasurementList(channelObj, 'DAQ', 'Event1', 'Measurement1');
    %       viewMeasurementLists(channelObj);
    %
    %   See also VNT.
    
        % Perform an argument count check.
        narginchk(1, 1);

        % Validate current state conditions.
        obj.errorWhenNoMeasurementListsConfigured();
        
        % Put the measurement lists together to configure them.
        measurementLists = [obj.DAQListObject obj.STIMListObject];

        % Display information for all configured measurements lists.
        fprintf(1, '\n');
        for ii = 1:obj.NumberOfMeasurementLists
            % Display basic list information.
            fprintf(1, '%s List #%d', measurementLists(ii).Resource, ii);
            if measurementLists(ii).EventInfo.ChannelTimeCycleInSeconds == 0
                fprintf(1, ' using the "%s" event ', measurementLists(ii).EventInfo.Name);
            else
                fprintf(1, ' using the "%s" event @ %f seconds ', ...
                    measurementLists(ii).EventInfo.Name, ...
                    measurementLists(ii).EventInfo.ChannelTimeCycleInSeconds);
            end
            fprintf(1, 'and the following measurements:\n');
            
            % Construct a disp table to display the measurements.
            measurementTable = internal.DispTable();
            measurementTable.ColumnSeparator = ' | ';
            measurementTable.ShowHeader = false;
            measurementTable.Indent = 3;
            measurementTable.addColumn('Measurements');
            for jj = 1:measurementLists(ii).NumberOfMeasurements
                measurementTable.addRow(measurementLists(ii).MeasurementNames{jj});
            end
            disp(measurementTable);
            
            % Display extra white space.
            fprintf(1, '\n');
        end
    end
    
    
    function startMeasurement(obj)
    % startMeasurement Start all configured DAQ and STIM lists.
    %
    %   startMeasurement(OBJ) actively starts all configured DAQ and STIM
    %   lists on the XCP channel object OBJ. DAQ lists begin acquiring
    %   data values from the server module. STIM lists begin transmitting
    %   data values to the server module.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %       createMeasurementList(channelObj, 'DAQ', 'Event1', 'Measurement1');
    %       startMeasurement(channelObj);
    %
    %   See also VNT.
    
        % Perform an argument count check.
        narginchk(1, 1);
        
        % Validate current state conditions.
        obj.errorWhenDisconnected();
        obj.errorWhenMeasurementActive();
        obj.errorWhenNoMeasurementListsConfigured();
        
        % Clear existing DAQ data.
        obj.DAQData = containers.Map;
        
        % Put the measurement lists together to configure them.
        measurementLists = [obj.DAQListObject obj.STIMListObject];
        
        % Free all existing measurement lists.
        obj.ProtocolObject.freeDAQ();
        
        % Configure for the required number of lists.
        obj.ProtocolObject.allocDAQ(obj.NumberOfMeasurementLists);
        
        % Configure the ODTs for each list. All ODTs must be
        % configured before continuing to add ODT entries and
        % measurements.
        for ii = 1:obj.NumberOfMeasurementLists
            % Configure this list for the required number of ODTs.
            obj.ProtocolObject.allocODT( ...
                ii - 1, ... % List number adjusted for zero index.
                measurementLists(ii).NumberOfODTs);
        end
        
        % Configure the ODT entries for each ODT on each list. All
        % ODT entries must be configured before measurements are added
        % to the entries.
        for ii = 1:obj.NumberOfMeasurementLists
            % Loop to configure the ODT entries for each ODT.
            for jj = 1:measurementLists(ii).NumberOfODTs
                % Configure the ODT entries.
                obj.ProtocolObject.allocODTEntry( ...
                    ii - 1, ... % List number adjusted for zero index.
                    jj - 1, ... % ODT number adjusted for zero index.
                    measurementLists(ii).ODTObject(jj).NumberOfODTEntries);
            end
        end
        
        % Add the measurements to the appropriate list, ODT, and
        % ODT entry.
        for ii = 1:obj.NumberOfMeasurementLists
            % Loop through each ODT to add measurements.
            for jj = 1:measurementLists(ii).NumberOfODTs
                % Set the DAQ pointer to prepare for linking the
                % measurement information with the DAQ list, ODT, and
                % ODT entry.
                obj.ProtocolObject.setDAQPtr( ...
                    ii - 1, ... % List number adjusted for zero index.
                    jj - 1, ... % ODT number adjusted for zero index.
                    xcp.Utility.FIRST_ODT_ENTRY);
                
                % Loop to write each measurement to their ODT entries.
                for kk = 1:measurementLists(ii).ODTObject(jj).NumberOfODTEntries
                    % Write the measurement information to the ODT.
                    obj.ProtocolObject.writeDAQ( ...
                        xcp.Utility.IGNORE_BIT_OFFSET, ...
                        measurementLists(ii).ODTObject(jj).ODTEntryInfo(kk).sizeInBytes(), ...
                        measurementLists(ii).ODTObject(jj).ODTEntryInfo(kk).ECUAddressExtension, ...
                        measurementLists(ii).ODTObject(jj).ODTEntryInfo(kk).ECUAddress);

                    % If this measurement list is a DAQ list, then add 
                    % an entry for this measurement to the DAQ data.
                    if strcmpi(measurementLists(ii).Resource, 'DAQ')
                        obj.DAQData(measurementLists(ii).ODTObject(jj).ODTEntryInfo(kk).Name) = [];
                        iival= int32(ii);
                        jjval= int32(jj);
                        kkval= int32(kk);
                        odtEntrySize = int32(measurementLists(ii).ODTObject(jj).ODTEntryInfo(kk).sizeInBytes());
                        listMode = typecast(measurementLists(ii).ListMode,'uint8');
                        mname=measurementLists(ii).ODTObject(jj).ODTEntryInfo(kk).Name;
                        obj.api_xcp_daqObject.addMeasurement( mname,iival-1,jjval-1,kkval-1,odtEntrySize,listMode);
        		locdatatype = int32(measurementLists(ii).ODTObject(jj).ODTEntryInfo(kk).LocDataType);
        		convtype = int32(measurementLists(ii).ODTObject(jj).ODTEntryInfo(kk).Conversion.ConversionType);
                        byteOrder= int32(measurementLists(ii).ODTObject(jj).ODTEntryInfo(kk).ByteOrder);
	                obj.api_xcp_daqObject.addComputeMethod(mname,iival-1,locdatatype, ...
                        measurementLists(ii).ODTObject(jj).ODTEntryInfo(kk).Conversion.a,...
			        measurementLists(ii).ODTObject(jj).ODTEntryInfo(kk).Conversion.b,convtype,byteOrder, ...
            measurementLists(ii).ODTObject(jj).ODTEntryInfo(kk).LowerLimit, ...
            measurementLists(ii).ODTObject(jj).ODTEntryInfo(kk).UpperLimit);
                    end
                end
            end
        end
        
        % Configure the lists for starting.
        for ii = 1:obj.NumberOfMeasurementLists
            % Set the measurement list mode.
            obj.ProtocolObject.setDAQListMode( ...
                measurementLists(ii).ListMode, ...
                ii - 1, ... % List number adjusted for zero index.
                measurementLists(ii).EventInfo.ChannelNumber, ...
                xcp.Utility.NO_TRANSMISSION_RATE_PRESCALER, ...
                xcp.Utility.LET_SERVER_MANAGE_PRIORITY);

            % Prepare the list to start.
            [~, firstNumber] = obj.ProtocolObject.startStopDAQList( ...
                xcp.Utility.SET_SYNCHRONIZED_START_STOP, ...
                ii - 1); % List number adjusted for zero index.

            % Set the absolute list number for the list.
            measurementLists(ii).Number = ii - 1;
            % Set the numbers for the ODTs in the list.
            measurementLists(ii).setODTNumbers(firstNumber);
        end

        % Configure STIM messages for each STIM list.
        for ii = 1:obj.NumberOfSTIMLists
            obj.STIMListObject(ii).configureSTIMMessageObjects();
        end
        
        % Set Flag to enable DAQ Processing  for CyFlex 
	enableDAQProcessing(obj);
        try
            % Turn on the lists.
            obj.ProtocolObject.startStopSynch(xcp.Utility.SYNCH_START_SELECTED_LISTS);
        catch ME
            if obj.ProtocolObject.ReceiverOverrun
                error(message('vnt:XCP:IncomingDataTooFast', obj.ProtocolObject.TransportLayerObject.BytesLost));
            else
                rethrow(ME);
            end
        end
        
        % Configure and start a timer to process DAQ messages if DAQ is
        % being used as a resource.
        %%if ~isempty(obj.DAQListObject)
        %%    obj.DAQTimer = internal.IntervalTimer(obj.DAQTimerRate);
        %%    obj.DAQTimerListener = event.listener( ...
        %%        obj.DAQTimer, ...
        %%        'Executing', ...
         %%       @(~, ~)(obj.executeDAQTimerCallback));
        %%    start(obj.DAQTimer);
        %%end

        % Configure and start a timer to process STIM messages if STIM is
        % being used as a resource.
        if ~isempty(obj.STIMListObject)
            obj.STIMTimer = internal.IntervalTimer(obj.STIMTimerRate);
            obj.STIMTimerListener = event.listener( ...
                obj.STIMTimer, ...
                'Executing', ...
                @(~, ~)(obj.executeSTIMTimerCallback));
            start(obj.STIMTimer);
        end
        
        % Modify the internal state.
        obj.MeasurementState = true;
    end
    
    function stopMeasurement(obj)
    % stopMeasurement Stop all configured DAQ and STIM lists.
    %
    %   stopMeasurement(OBJ) stops all configured DAQ and STIM
    %   lists on the XCP channel object OBJ. DAQ lists stop acquiring
    %   data values from the server module. STIM lists stop transmitting
    %   data values to the server module.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %       createMeasurementList(channelObj, 'DAQ', 'Event1', 'Measurement1');
    %       startMeasurement(channelObj);
    %       stopMeasurement(channelObj);
    %
    %   See also VNT.
    
        % Perform an argument count check.
        narginchk(1, 1);

        % Validate current state conditions.
        obj.errorWhenDisconnected();
        obj.errorWhenMeasurementInactive();
        
        % Stop and remove the STIM timer if in use.
        if ~isempty(obj.STIMTimer)
            stop(obj.STIMTimer);
            obj.STIMTimer = [];
        end
        
        % Turn off all of the lists.
        obj.ProtocolObject.startStopSynch(xcp.Utility.SYNCH_STOP_ALL_LISTS);
        % set flag to disable  processing DAQ FRAME for CyFlex
         disableDAQProcessing(obj);

        % Stop and remove the DAQ timer if in use.
        if ~isempty(obj.DAQTimer)
            stop(obj.DAQTimer);
            obj.DAQTimer = [];
        end
        
        % Modify the internal state.
        obj.MeasurementState = false;
    end
    
    
    function data = readDAQListData(obj, m, varargin)
    % readDAQListData Read samples of the specified measurement from a DAQ list.
    %
    %   DATA = readDAQListData(OBJ, MEASUREMENT) returns all 
    %   acquired DAQ list data from the XCP channel object OBJ for the 
    %   specified MEASUREMENT. 
    %
    %   DATA = readDAQListData(OBJ, MEASUREMENT, COUNT) returns 
    %   acquired DAQ list data from the XCP channel object OBJ for the 
    %   specified MEASUREMENT. In this case, the quantity of data
    %   values to read is given as COUNT. That number of values will be
    %   returned or less if fewer samples are available.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %       createMeasurementList(channelObj, 'DAQ', 'Event1', 'Measurement1');
    %       startMeasurement(channelObj);
    %       data = readDAQListData(channelObj, 'Measurement1', 10);
    %       data = readDAQListData(channelObj, 'Measurement1');
    %
    %   See also VNT.
        
        % Perform an argument count check.
        narginchk(2, 3);

        % Convert string inputs to character vectors.
        m = convertStringsToChars(m);

        % Validate the measurement argument.
        validateattributes(m, {'xcp.a2l.Measurement', 'char'}, {'nonempty'},...
            'xcp.readDAQListData', 'measurement', 2);

        % Get the detailed information for this measurement.
        if ischar(m)
            try
                m = obj.A2LFileObject.MeasurementInfo(m);
            catch
                % Rethrow not found error as a more usable error message to
                % the user. The default containers.Map error is not helpful.
                error(message('vnt:XCP:ElementNotFound', 'measurement', m));
            end
        end

        % Validate the number of samples argument if it exists.
        if nargin == 3
            validateattributes(varargin{1}, {'numeric'}, ...
                {'nonempty', 'finite', 'nonnan', 'positive', 'integer', 'real'}, ...
                'xcp.readDAQListData', 'count', 3);
        end
        
        % If the internal DAQ data is empty, then return empty as no data
        % is yet available.
        if isempty(obj.DAQData)
            data = [];
            return;
        end
        
        % If the measurement is not in the DAQ data, then it has not
        % been added to a DAQ list, so error.
        if ~obj.DAQData.isKey(m.Name)
            error(message('vnt:XCP:MeasurementNotConfiguredForDAQ', m.Name));
        end
        
        % Read the number of samples currently available.
        daqValue = obj.DAQData(m.Name);
        sampleCount = numel(daqValue);
        
        % If no samples are available, then return empty.
        if sampleCount == 0
            data = [];
            return;
        end
        
        % If a sample count was provided and that value is less than the
        % number of samples available, then read only that quantity.
        if nargin == 3 && varargin{1} < sampleCount
            sampleCount = varargin{1};
        end
        
        % Return the data.
        data = daqValue(1:sampleCount);
        
        % Clear the read data from the object.
        daqValue(1:sampleCount) = [];
        obj.DAQData(m.Name) = daqValue;
    end
    
    function writeSTIMListData(obj, m, newValue)
    % writeSTIMListData Write a new value of the specified measurement to a STIM list.
    %
    %   writeSTIMListData(OBJ, MEASUREMENT, NEW_VALUE) writes a
    %   NEW_VALUE to the STIM list measurement MEASUREMENT on the XCP
    %   channel object OBJ.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %       createMeasurementList(channelObj, 'STIM', 'Event1', 'Measurement1');
    %       startMeasurement(channelObj);
    %       writeSTIMListData(channelObj, 'Measurement1', newValue);
    %
    %   See also VNT.

        % Perform an argument count check.
        narginchk(3, 3);        

        % Convert string inputs to character vectors.
        m = convertStringsToChars(m);

        % Validate the measurement argument.
        validateattributes(m, {'xcp.a2l.Measurement', 'char'}, {'nonempty'},...
            'xcp.writeSTIMListData', 'measurement', 2);

        % Get the detailed information for this measurement.
        if ischar(m)
            try
                m = obj.A2LFileObject.MeasurementInfo(m);
            catch
                % Rethrow not found error as a more usable error message to
                % the user. The default containers.Map error is not helpful.
                error(message('vnt:XCP:ElementNotFound', 'measurement', m));
            end
        end
        
        % Validate the value argument.
        validateattributes(newValue, {'numeric'}, ...
            {'nonempty', 'finite', 'nonnan', 'real'}, ...
            'xcp.writeSTIMListData', 'new value', 3);
        
        % Pack the data value.
        for ii = 1:obj.NumberOfSTIMLists
            % Try to write the value to each STIM list.
            result = obj.STIMListObject(ii).writeSTIMData(m, newValue, false);
            % Check the result. If the write worked, then return.
            if result
                return;
            end
        end
        
        % If no write was made, error as the measurement was not found to
        % be configured for STIM.
        error(message('vnt:XCP:MeasurementNotConfiguredForSTIM', m.Name));
    end

    
    function data = readDAQ(obj, m, varargin)
    % readDAQ Read scaled samples of the specified measurement from a DAQ list.
    %
    %   DATA = readDAQ(OBJ, MEASUREMENT) returns all acquired DAQ list 
    %   data from the XCP channel object OBJ for the specified MEASUREMENT. 
    %
    %   DATA = readDAQ(OBJ, MEASUREMENT, COUNT) returns acquired DAQ list 
    %   data from the XCP channel object OBJ for the specified MEASUREMENT. 
    %   In this case, the quantity of data values to read is given as COUNT. 
    %   That number of values will be returned or less if fewer samples 
    %   are available.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %       createMeasurementList(channelObj, 'DAQ', 'Event1', 'Measurement1');
    %       startMeasurement(channelObj);
    %       data = readDAQ(channelObj, 'Measurement1', 10);
    %       data = readDAQ(channelObj, 'Measurement1');
    %
    %   See also VNT.
        
        % Perform an argument count check.
        narginchk(2, 3);

        % Convert string inputs to character vectors.
        m = convertStringsToChars(m);

        % Validate the measurement argument.
        validateattributes(m, {'xcp.a2l.Measurement', 'char'}, {'nonempty'},...
            'xcp.readDAQ', 'measurement', 2);

        % Get the detailed information for this measurement.
        if ischar(m)
            try
                m = obj.A2LFileObject.MeasurementInfo(m);
            catch
                % Rethrow not found error as a more usable error message to
                % the user. The default containers.Map error is not helpful.
                error(message('vnt:XCP:ElementNotFound', 'measurement', m));
            end
        end
        
        % Validate the number of samples argument if it exists.
        if nargin == 3
            validateattributes(varargin{1}, {'numeric'}, ...
                {'nonempty', 'finite', 'nonnan', 'positive', 'integer', 'real'}, ...
                'xcp.readDAQ', 'count', 3);
        end
        
        % Read the raw value of the data.
        data = obj.readDAQListData(m, varargin{:});
        
        % Convert to scaled values.
        data = m.convertToPhy(data);
        
         % Reshape
        data = reshape(data, [m.Dimension length(data)/prod(m.Dimension)]);
    end

    function writeSTIM(obj, m, newValue)
    % writeSTIM Write a scaled new value of the specified measurement to a STIM list.
    %
    %   writeSTIM(OBJ, MEASUREMENT, NEW_VALUE) writes a NEW_VALUE to the 
    %   STIM list measurement MEASUREMENT on the XCP channel object OBJ.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %       createMeasurementList(channelObj, 'STIM', 'Event1', 'Measurement1');
    %       startMeasurement(channelObj);
    %       writeSTIM(channelObj, 'Measurement1', newValue);
    %
    %   See also VNT.
        
        % Perform an argument count check.
        narginchk(3, 3);

        % Convert string inputs to character vectors.
        m = convertStringsToChars(m);

        % Validate the measurement argument.
        validateattributes(m, {'xcp.a2l.Measurement', 'char'}, {'nonempty'},...
            'xcp.writeSTIM', 'measurement', 2);

        % Get the detailed information for this measurement.
        if ischar(m)
            try
                m = obj.A2LFileObject.MeasurementInfo(m);
            catch
                % Rethrow not found error as a more usable error message to
                % the user. The default containers.Map error is not helpful.
                error(message('vnt:XCP:ElementNotFound', 'measurement', m));
            end
        end

        % Validate the value argument.
        validateattributes(newValue, {'numeric'}, ...
            {'nonempty', 'finite', 'nonnan', 'real'}, ...
            'xcp.writeSTIM', 'new value', 3);

        % Pack the data value.
        for ii = 1:obj.NumberOfSTIMLists
            % Try to write the value to each STIM list.
            result = obj.STIMListObject(ii).writeSTIMData(m, newValue, true);
            % Check the result. If the write worked, then return.
            if result
                return;
            end
        end

        % If no write was made, error as the measurement was not found to
        % be configured for STIM.
        error(message('vnt:XCP:MeasurementNotConfiguredForSTIM', m.Name));
    end
    
    
    function value = readSingleValue(obj, m)
    % readSingleValue Read a sample of the specified measurement from direct memory.
    %
    %   VALUE = readSingleValue(OBJ, MEASUREMENT) acquires a single
    %   VALUE for the specified MEASUREMENT through the XCP channel
    %   object OBJ. This action is performed via direct read from memory on
    %   the server module.
    %
    %   Consider using readMeasurement instead, especially if you want
    %   scaled values or have non-native type and non-scalar data.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %       value = readSingleValue(channelObj, 'Measurement1');
    %
    %   See also VNT.
    
        % Perform an argument count check.
        narginchk(2, 2);

        % Convert string inputs to character vectors.
        m = convertStringsToChars(m);
        
        % Validate the measurement argument.
        validateattributes(m, {'xcp.a2l.Measurement', 'char'}, {'nonempty'},...
            'xcp.readSingleValue', 'measurement', 2);
        
        % Get the detailed information for this measurement.
        if ischar(m)
            try
                m = obj.A2LFileObject.MeasurementInfo(m);
            catch
                % Rethrow not found error as a more usable error message to
                % the user. The default containers.Map error is not helpful.
                error(message('vnt:XCP:ElementNotFound', 'measurement', m));
            end
        end
        
        % Validate current state conditions.
        obj.errorWhenDisconnected();

        % Set the memory transfer address for this measurement.
        obj.ProtocolObject.setMTA(m.ECUAddressExtension, m.ECUAddress);
        
        % Upload the data value from the server.
        response = obj.ProtocolObject.upload(m.sizeInBytes());

        % Decode and return the measurement value.
        firstByte = 2;
        lastByte = firstByte + m.sizeInBytes() - 1;
        value = xcp.Utility.unpackMeasurement(m, response(firstByte:lastByte));
    end
    
    function writeSingleValue(obj, m, newValue)
    % writeSingleValue Write a new value of the specified measurement to direct memory.
    %
    %   writeSingleValue(OBJ, MEASUREMENT_NAME, NEW_VALUE) writes a single
    %   NEW_VALUE for the specified MEASUREMENT_NAME through the XCP channel
    %   object OBJ. This action is performed via direct write to memory on
    %   the server module.
    %
    %   Consider using writeMeasurement instead, especially if you want
    %   scaled values or have non-native type and non-scalar data.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %       writeSingleValue(channelObj, 'Measurement1', newValue);
    %
    %   See also VNT.
    
        % Perform an argument count check.
        narginchk(3, 3);

        % Convert string inputs to character vectors.
        m = convertStringsToChars(m);
        
        % Validate the measurement argument.
        validateattributes(m, {'xcp.a2l.Measurement', 'char'}, {'nonempty'}, ...
            'xcp.writeSingleValue', 'measurement', 2);
        
        % Get the detailed information for this measurement.
        if ischar(m)
            try
                m = obj.A2LFileObject.MeasurementInfo(m);
            catch
                % Rethrow not found error as a more usable error message to
                % the user. The default containers.Map error is not helpful.
                error(message('vnt:XCP:ElementNotFound', 'measurement', m));
            end
        end

        % Validate the value argument.
        validateattributes(newValue, {'numeric'}, {'nonempty', 'real', 'finite', 'nonnan'}, ...
            'xcp.writeSingleValue', 'new value', 3);

        % Validate current state conditions.
        obj.errorWhenDisconnected();

        % Convert the data value into bytes suitable for download.
        byteValue = xcp.Utility.packMeasurement(m, newValue);

        % Set the memory transfer address for this measurement.
        obj.ProtocolObject.setMTA(m.ECUAddressExtension, m.ECUAddress);
        % Download the bytes to the server.
        obj.ProtocolObject.download(m.sizeInBytes(), byteValue);
    end

    function value = readMeasurement(obj, m)
    % readMeasurement Read a scaled sample of the specified measurement from direct memory.
    %
    %   VALUE = readMeasurement(OBJ, MEASUREMENT) acquires and scales the
    %   VALUE for the specified MEASUREMENT through the XCP channel
    %   object OBJ. This action is performed via direct read from memory on
    %   the server module. MEASUREMENT can be either an object obtained from
    %   the A2L file interface or the name of the measurement as a string.
    %
    %   Example:
    %       a2lObj = xcp.A2L('myFile.a2l');
    %       channelObj = xcp.Channel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %
    %       value = readMeasurement(channelObj, 'myMeasurement');
    %
    %       measurementObj = a2lObj.MeasurementInfo('myMeasurement');
    %       value = readMeasurement(channelObj, measurementObj);
    %
    %   See also VNT.
        
        % Perform an argument count check.
        narginchk(2, 2);
        
        % Convert string inputs to character vectors.
        m = convertStringsToChars(m);
        
        % Validate the measurement argument.
        validateattributes(m, {'xcp.a2l.Measurement', 'char'}, {'nonempty'},...
            'xcp.readMeasurement', 'measurement', 2);
        
        % Get the detailed information for this measurement.
        if ischar(m)
            try
                m = obj.A2LFileObject.MeasurementInfo(m);
            catch
                % Rethrow not found error as a more usable error message to
                % the user. The default containers.Map error is not helpful.
                error(message('vnt:XCP:ElementNotFound', 'measurement', m));
            end
        end
        
        % Read and scale the measurement.
        frame = obj.readFrame(m.ECUAddress, m.ECUAddressExtension, m.sizeInBytes());
        value = m.convertToPhy(frame);
    end
    
    function writeMeasurement(obj, m, newValue)
    % writeMeasurement Write a new scaled value of the specified measurement to direct memory.
    %
    %   writeMeasurement(OBJ, MEASUREMENT, NEW_VALUE) scales and writes a
    %   NEW_VALUE for the specified MEASUREMENT through the XCP channel
    %   object OBJ. This action is performed via direct write to memory on
    %   the server module. MEASUREMENT can be either an object obtained from
    %   the A2L file interface or the name of the measurement as a string.
    %
    %   Example:
    %       a2lObj = xcp.A2L('myFile.a2l');
    %       channelObj = xcp.Channel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %
    %       writeMeasurement(channelObj, 'myMeasurement', newValue);
    %
    %       measurementObj = a2lObj.MeasurementInfo('myMeasurement');
    %       writeMeasurement(channelObj, measurementObj, newValue);
    %
    %   See also VNT.
        
        % Perform an argument count check.
        narginchk(3, 3);
        
        % Convert string inputs to character vectors.
        m = convertStringsToChars(m);
        
        % Validate the measurement argument.
        validateattributes(m, {'xcp.a2l.Measurement', 'char'}, {'nonempty'}, ...
            'xcp.writeMeasurement', 'measurement', 2);
        
        % Get the detailed information for this measurement.
        if ischar(m)
            try
                m = obj.A2LFileObject.MeasurementInfo(m);
            catch
                % Rethrow not found error as a more usable error message to
                % the user. The default containers.Map error is not helpful.
                error(message('vnt:XCP:ElementNotFound', 'measurement', m));
            end
        end
        
        % Validate the value argument.
        validateattributes(newValue, {'numeric', 'char', 'cell' }, ...
            {'nonempty'}, 'xcp.writeMeasurement', 'new value', 3);        
        
        % Verify the value is properly sized for this measurement.
        cm = m.conversion();
        if m.dimension() ~= numel(newValue) && ~isempty(cm) && ~(isa(cm, 'xcp.a2l.CompuMethodTable') && strcmp(cm.ConversionType,'TAB_VERB'))
            error(message('asam_a2l:messages:IncorrectValueSize', m.Name, num2str(m.dimension()')));
        end
        
        % Scale and write the measurement.
        frame = m.convertToRaw(newValue);
        obj.writeFrame(frame, m.ECUAddress, m.ECUAddressExtension, m.sizeInBytes());
    end
    
    function value = readCharacteristic(obj, c, varargin)
    % readCharacteristic Read a scaled sample of the specified characteristic from direct memory.
    %
    %   VALUE = readCharacteristic(OBJ, CHARACTERISTIC) reads and scales a
    %   VALUE for the specified CHARACTERISTIC through the XCP channel
    %   object OBJ. This action is performed via direct read from memory on
    %   the server module. CHARACTERISTIC can be either an object obtained from
    %   the A2L file interface or the name of the measurement as a string.
    %
    %   VALUE = readCharacteristic(OBJ, CHARACTERISTIC, ELEMENT) reads and scales a
    %   VALUE for the specified CHARACTERISTIC through the XCP channel
    %   object OBJ. ELEMENT is a string allowing selection of which part of
    %   the CHARACTERISTIC to read. It is specified as 'AXIS_PTS_X',
    %   'AXIS_PTS_Y', 'AXIS_PTS_Z', or 'FNC_VALUES'. When ELEMENT is note
    %   provided, 'FNC_VALUES' is used.
    %
    %   Example:
    %       a2lObj = xcp.A2L('myFile.a2l');
    %       channelObj = xcp.Channel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %
    %       value = readCharacteristic(channelObj, 'myCharacteristic');
    %
    %       characteristicObj = a2lObj.CharacteristicInfo('myCharacteristic');
    %       value = readCharacteristic(channelObj, characteristicObj);
    %
    %       characteristicObj = a2lObj.CharacteristicInfo('myCharacteristic');
    %       value = readCharacteristic(channelObj, characteristicObj, 'AXIS_PTS_X');
    %
    %   See also VNT.
        
        % Perform an argument count check.
        narginchk(2, 3);
        
        % Convert string inputs to character vectors.
        c = convertStringsToChars(c);
        
        % Validate the characteristic argument.
        validateattributes(c, {'xcp.a2l.Characteristic', 'char'}, {'nonempty'}, ...
            'xcp.readCharacteristic', 'characteristic', 2);
        
        % Get the detailed information for this characteristic.
        if ischar(c)
            try
                c = obj.A2LFileObject.CharacteristicInfo(c);
            catch
                % Rethrow not found error as a more usable error message to
                % the user. The default containers.Map error is not helpful.
                error(message('vnt:XCP:ElementNotFound', 'characteristic', c));
            end
        end
        
        % Determine the target element of the characteristic.
        if nargin == 3
            % Validate the element argument.
            element = validatestring(varargin{1}, {'AXIS_PTS_X', 'AXIS_PTS_Y', 'AXIS_PTS_Z', 'FNC_VALUES'}, ...
                'xcp.readCharacteristic', 'element', 3);            
        else
            % When unspecified, we read FNC_VALUES.
            element = 'FNC_VALUES';
        end
        
        % Determine needed detailed about the characteristic.
        [offset, sizeb] = getLayoutOffsetAndSize(c, element);
        address = c.ECUAddress + offset;
    
        % Read and scale the characteristic.
        frame = obj.readFrame(address, c.ECUAddressExtension, sizeb);
        value = c.convertToPhy(frame, element);
    end
    
    function writeCharacteristic(obj, c, newValue, varargin)
    % writeCharacteristic Write a new scaled value of the specified characteristic to direct memory.
    %
    %   writeCharacteristic(OBJ, CHARACTERISTIC, NEW_VALUE) scales and writes a
    %   NEW_VALUE for the specified CHARACTERISTIC through the XCP channel
    %   object OBJ. This action is performed via direct write to memory on
    %   the server module. CHARACTERISTIC can be either an object obtained from
    %   the A2L file interface or the name of the measurement as a string.
    %
    %   Example:
    %       a2lObj = xcp.A2L('myFile.a2l');
    %       channelObj = xcp.Channel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %
    %       writeCharacteristic(channelObj, 'myCharacteristic', newValue);
    %
    %       characteristicObj = a2lObj.CharacteristicInfo('myCharacteristic');
    %       writeCharacteristic(channelObj, characteristicObj, newValue);
    %
    %       characteristicObj = a2lObj.CharacteristicInfo('myCharacteristic');
    %       writeCharacteristic(channelObj, characteristicObj, newValue, 'AXIS_PTS_X');
    %
    %   See also VNT.
        
        % Perform an argument count check.
        narginchk(3, 4);
        
        % Convert string inputs to character vectors.
        c = convertStringsToChars(c);
        
        % Validate the characteristic argument.
        validateattributes(c, {'xcp.a2l.Characteristic', 'char'}, {'nonempty'}, ...
            'xcp.writeCharacteristic', 'characteristic', 2);
        
        % Get the detailed information for this characteristic.
        if ischar(c)
            try
                c = obj.A2LFileObject.CharacteristicInfo(c);
            catch
                % Rethrow not found error as a more usable error message to
                % the user. The default containers.Map error is not helpful.
                error(message('vnt:XCP:ElementNotFound', 'characteristic', c));
            end
        end
        
        % Validate the value argument.
        validateattributes(newValue, {'numeric', 'char', 'cell' }, ...
            {'nonempty'}, 'xcp.writeCharacteristic', 'new value', 3);
        
        % Determine the target element of the characteristic. 
        if nargin == 4
            % Validate the element argument.
            element = validatestring(varargin{1}, {'AXIS_PTS_X', 'AXIS_PTS_Y', 'AXIS_PTS_Z', 'FNC_VALUES'}, ...
                'xcp.writeCharacteristic', 'element', 3);            
        else
            % When unspecified, we write FNC_VALUES.
            element = 'FNC_VALUES';
        end
        
        % Determine needed detailed about the characteristic.
        [offset, size, dims] = getLayoutOffsetAndSize(c, element);
        address = c.ECUAddress + offset;

        % Verify the value is properly sized for this measurement.
        cm = c.conversion();
        if prod(dims) ~= numel(newValue) && ~isempty(cm) && ~(isa(cm, 'xcp.a2l.CompuMethodTable') && strcmp(cm.ConversionType,'TAB_VERB'))
            error(message('asam_a2l:messages:IncorrectValueSize', c.Name, num2str(prod(dims))));
        end

        % Scale and write the characteristic.
        frame = c.convertToRaw(newValue, element);
        obj.writeFrame(frame, address, c.ECUAddressExtension, size);
    end


    function value = readAxis(obj, a)
    % readAxis Read a scaled sample of the specified axis from direct memory.
    %
    %   VALUE = readAxis(OBJ, AXIS) read and scales a VALUE for the 
    %   specified AXIS through the XCP channel object OBJ. This action is 
    %   performed via direct read from memory on the server module. AXIS can 
    %   be either an object obtained from the A2L file interface or the 
    %   name of the measurement as a string.
    %
    %   Example:
    %       a2lObj = xcp.A2L('myFile.a2l');
    %       channelObj = xcp.Channel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %
    %       value = readAxis(channelObj, 'myAxis');
    %
    %       axisObj = a2lObj.AxisInfo('myAxis');
    %       value = readAxis(channelObj, axisObj);
    %
    %   See also VNT.
        
        % Perform an argument count check.
        narginchk(2, 2);
        
        % Convert string inputs to character vectors.
        a = convertStringsToChars(a);
        
        % Validate the axis argument.
        validateattributes(a, {'xcp.a2l.AxisPts', 'char'}, {'nonempty'}, ...
            'xcp.readAxis', 'axis', 2);
        
        % Get the detailed information for this axis.
        if ischar(a)
            try
                a = obj.A2LFileObject.AxisInfo(a);
            catch
                % Rethrow not found error as a more usable error message to
                % the user. The default containers.Map error is not helpful.
                error(message('vnt:XCP:ElementNotFound', 'axis', a));
            end
        end

        % Read and scale the axis.
        frame = obj.readFrame(a.ECUAddress, a.ECUAddressExtension, a.sizeInBytes());
        value = a.convertToPhy(frame);
    end

    function writeAxis(obj, a, newValue)
    % writeAxis Write a new scaled value of the specified axis to direct memory.
    %
    %   writeAxis(OBJ, AXIS, NEW_VALUE) scales and writes a NEW_VALUE for 
    %   the specified AXIS_OBJ through the XCP channel object OBJ. This 
    %   action is performed via direct write to memory on the server module.
    %   AXIS can be either an object obtained from the A2L file interface 
    %   or the name of the measurement as a string.
    %
    %   Example:
    %       a2lObj = xcp.A2L('myFile.a2l');
    %       channelObj = xcp.Channel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %
    %       writeAxis(channelObj, axisObj, 'myAxis');
    %
    %       axisObj = a2lObj.AxisInfo('myAxis');
    %       writeAxis(channelObj, axisObj, newValue);
    %
    %   See also VNT.
        
        % Perform an argument count check.
        narginchk(3, 3);
        
        % Convert string inputs to character vectors.
        a = convertStringsToChars(a);
        
        % Validate the axis argument.
        validateattributes(a, {'xcp.a2l.AxisPts','char'}, {'nonempty'}, ...
            'xcp.writeAxis', 'axis', 2);
        
       % Get the detailed information for this axis.
       if ischar(a)
            try
                a = obj.A2LFileObject.AxisInfo(a);
            catch
                % Rethrow not found error as a more usable error message to
                % the user. The default containers.Map error is not helpful.
                error(message('vnt:XCP:ElementNotFound', 'axis', a));
            end
        end
        
        % Validate the value argument.
        validateattributes(newValue, {'numeric', 'char', 'cell' }, ...
            {'nonempty'}, 'xcp.writeAxis', 'new value', 3);
        
        % Scale and write the axis.
        frame = a.convertToRaw(newValue);
        obj.writeFrame(frame, a.ECUAddress, a.ECUAddressExtension, a.sizeInBytes());
    end

    
    function result = isConnected(obj)
    % isConnected - Returns a boolean value to indicate active connection to the server.
    %
    %   RESULT = isConnected(OBJ) returns a boolean RESULT indicating true
    %   if the XCP channel object OBJ is actively connected to the server
    %   module, otherwise it returns false.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %       result = isConnected(channelObj);
    %
    %   See also VNT.
    
        % Return an indication based off the internal state property.
        result = obj.ConnectState;
    end
    
    function result = isMeasurementRunning(obj)
    % isMeasurementRunning - Returns a boolean value to indicate active measurement activity.
    %
    %   RESULT = isMeasurementRunning(OBJ) returns a boolean RESULT indicating
    %   true if the XCP channel object OBJ is actively running DAQ and/or STIM
    %   measurement lists with the server module, otherwise it returns false.
    %
    %   Example:
    %       a2lObj = xcpA2L('myFile.a2l');
    %       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'CANcaseXL 1', 1);
    %       connect(channelObj);
    %       createMeasurementList(channelObj, 'DAQ', 'Event1', 'Measurement1');
    %       startMeasurement(channelObj);
    %       result = isMeasurementRunning(channelObj);
    %
    %   See also VNT.
    
        % Return an indication based off the internal state property.
        result = obj.MeasurementState;
    end

    function out = startlog(obj) 
        out= obj.api_xcp_daqObject.startlog();
    end
    function stoplog( obj)
	 obj.api_xcp_daqObject.stoplog();
    end
  
end


methods (Access = 'private')

    function executeDAQTimerCallback(obj)
    % executeDAQTimerCallback A timer callback function that runs periodic DAQ tasks.
    
        % If there is no logged information at the protocol level, then exit.
       %% if numel(obj.ProtocolObject.DAQData) == 0
       %%     return;
      %%  end
        
        % Copy the information to a local variable to process it.
      %%  localData = obj.ProtocolObject.DAQData;
        % Delete the copied information in the protocol object.
      %%  obj.ProtocolObject.DAQData(1:numel(localData)) = [];
      %%  idFieldType =int32(obj.A2LFileObject.DAQInfo.IdentificationFieldType);
        
      %  for ii = 1:numel(localData)
     %   obj.api_xcp_daqObject.frameData(localData{1,ii}, idFieldType);
     %  end   
       return
        % Loop through and process each received entry.
        %for ii = 1:numel(localData)
            % Find the ODT in the DAQ lists that matches the ODT of the
            % received message.
        %    switch obj.A2LFileObject.DAQInfo.IdentificationFieldType
        %        case xcp.a2l.IdentificationField.IDENTIFICATION_FIELD_TYPE_ABSOLUTE
                    % When absolute ODT numbering is used, we not know the
                    % list number at this time, so we need to search
                    % through them all until we find the right ODT.
         %           for jj = 1:obj.NumberOfDAQLists
                        % Search the list for the ODT, and break when found.
          %              ODTObject = obj.DAQListObject(jj).getODTObject(localData{ii}(1));
           %             if ~isempty(ODTObject)
           %                 break;
           %             end
           %         end
                    
          %      case xcp.a2l.IdentificationField.IDENTIFICATION_FIELD_TYPE_RELATIVE_BYTE
                    % For relative byte identification fields, we know the
                    % list number and the relative ODT number. Both are
                    % byte values.
          %          ODTObject = obj.DAQListObject(1 + localData{ii}(2)).getODTObject(localData{ii}(1));
                    
          %      case xcp.a2l.IdentificationField.IDENTIFICATION_FIELD_TYPE_RELATIVE_WORD
                    % For relative word identification fields, again we
                    % know both the list number and relative ODT number.
                    % The ODT number is a byte. The list number is a word.
           %         listNumber = 1 + xcp.Utility.unpackParameter( ...
           %             obj.A2LFileObject.ProtocolLayerInfo.ByteOrder, ...
           %             xcp.Utility.BYTES_IN_WORD, ...
           %             localData{ii}(2:3));
           %         ODTObject = obj.DAQListObject(listNumber).getODTObject(localData{ii}(1));
                    
          %      case xcp.a2l.IdentificationField.IDENTIFICATION_FIELD_TYPE_RELATIVE_WORD_ALIGNED
                    % For relative word aligned identification fields, again we
                    % know both the list number and relative ODT number.
                    % The ODT number is a byte. The list number is a word,
                    % but it is located in different byte positions.
           %         listNumber = 1 + xcp.Utility.unpackParameter( ...
            %            obj.A2LFileObject.ProtocolLayerInfo.ByteOrder, ...
             %           xcp.Utility.BYTES_IN_WORD, ...
             %           localData{ii}(3:4));
             %       ODTObject = obj.DAQListObject(listNumber).getODTObject(localData{ii}(1));
          %  end
            
           % if ~isempty(ODTObject)
                % Set the byte index to begin reading data after the
                % identification field.
            %    currentByte = 2 + obj.A2LFileObject.DAQInfo.IdentificationFieldType;
             %   if ((localData{ii}(2) == 0) && (localData{ii}(1) == 0 ))           
             %      currentByte = 6 + obj.A2LFileObject.DAQInfo.IdentificationFieldType;
             %   end
                % Loop through and read all of the measurements in this ODT.
             %   for jj = 1:ODTObject.NumberOfODTEntries
                    % Use the measurement information to read the value.
              %      firstByte = currentByte;
              %      lastByte = firstByte + ODTObject.ODTEntryInfo(jj).sizeInBytes() - 1;
              %      data = xcp.Utility.unpackMeasurement(ODTObject.ODTEntryInfo(jj), localData{ii}(firstByte:lastByte));
                    
                    % Set this new data into the DAQ data arrays.
               %     daqValue = obj.DAQData(ODTObject.ODTEntryInfo(jj).Name);
               %     daqValue = [daqValue data];
                    obj.DAQData(ODTObject.ODTEntryInfo(jj).Name) = daqValue;
                    
                    % Increment the current byte counter to index the next ODT
                    % Entry in the message data.
              %      currentByte = currentByte + ODTObject.ODTEntryInfo(jj).sizeInBytes();
              %  end
            % end
            
       % end
    end
    
    function executeSTIMTimerCallback(obj)
    % executeSTIMTimerCallback A timer callback function that runs periodic STIM tasks.

        % Loop through each STIM list.
        for ii = 1:obj.NumberOfSTIMLists
            % Subtract the STIM timer rate from the counter.
            obj.STIMListObject(ii).STIMCounter = obj.STIMListObject(ii).STIMCounter - obj.STIMTimerRate;
            
            % If the counter has gone negative then it is time for this
            % list to be transmitted.
            if obj.STIMListObject(ii).STIMCounter <= 0
                % Transmit the message for all ODTs in the list.
                obj.ProtocolObject.TransportLayerObject.transmit([obj.STIMListObject(ii).ODTObject.STIMMessageObject]);
                % Reset the counter.
                obj.STIMListObject(ii).STIMCounter = obj.STIMListObject(ii).STIMPeriod;
            end
        end
    end

    function unlockResource(obj, resource)
    % unlockResource Unlocks a server with seed and key security access.
    %
    %   resource - A string indicating which resource to unlock.
        
        % Check that a seed and key DLL file is configured.
        if isempty(obj.SeedKeyDLLFullPath)
            error(message('vnt:XCP:NoSeedKeyDLLFileConfigured', resource));
        end
        
        % Call for the first part of the seed.
        [~, seedLength, seedValue] = obj.ProtocolObject.getSeed( ...
            xcp.Utility.FIRST_PART_OF_SEED, ...
            xcp.Utility.(['RESOURCE_' resource]));
        % Store the number of seed bytes read so far.
        seedBytesRead = numel(seedValue);
        
        % Get the additional parts of the seed value if necessary.
        while seedBytesRead ~= seedLength
            % Call for additional seed bytes.
            [~, ~, seedValueAdditional] = obj.ProtocolObject.getSeed( ...
                xcp.Utility.REMAINING_PART_OF_SEED, ...
                xcp.Utility.(['RESOURCE_' resource]));
            % Put the seed bytes together.
            seedValue = [seedValue seedValueAdditional]; %#ok<AGROW>
            % Add to the number of seed bytes read.
            seedBytesRead = seedBytesRead + numel(seedValueAdditional);
        end
        
        % Call the seed and key DLL MEX interface to turn the seed into a key.
        [status, keyLength, keyValue] = mexXCPSeedKey( ...
            'skGetKeyFromSeed', ...
            obj.SeedKeyDLLFullPath, ...
            xcp.Utility.(['RESOURCE_' resource]), ...
            seedLength, ...
            seedValue);
        
        % Verify that the key value was obtained.
        if status ~= xcp.Utility.SEED_KEY_SUCCESS
            error(message('vnt:XCP:KeyGenerationFailed', resource, xcp.Utility.MAP_SEED_KEY_RESPONSE_TO_NAME(status)));
        end        
        
        % Use the key value to unlock the resource.
        keyBytesRemainingToSend = keyLength;
        while keyBytesRemainingToSend > 0
            % Determine which key value bytes to send.
            if keyBytesRemainingToSend <= xcp.Utility.MAX_KEY_BYTES_PER_CTO
                % Send all available bytes.
                keyBytesToSend = keyValue;
                % Clear the key value.
                keyValue = [];
            else
                % Send the maximum number of bytes that fit into the CTO.
                keyBytesToSend = keyValue(1:xcp.Utility.MAX_KEY_BYTES_PER_CTO);
                % Clear those bytes in the key value.
                keyValue(1:xcp.Utility.MAX_KEY_BYTES_PER_CTO) = [];
            end
                
            % Issue an unlock command.
            [~, protectionStatus] = obj.ProtocolObject.unlock(keyBytesRemainingToSend, keyBytesToSend);
            
            % Adjust the number of remaining key bytes to send.
            keyBytesRemainingToSend = keyBytesRemainingToSend - numel(keyBytesToSend);
        end
        
        % Verify that the resource was unlocked.
        if bitand(protectionStatus, xcp.Utility.(['RESOURCE_' resource]))
            error(message('vnt:XCP:ResourceDidNotUnlock', resource));
        end
    end
    
    
    function errorWhenConnected(obj)
    % errorWhenConnected Throws an error if the channel is connected.
    
        % Check the state.
        if obj.isConnected()
            % Throw an error to prevent operation by the calling method.
            error(message('vnt:XCP:ChannelConnected'));
        end
    end
    
    function errorWhenDisconnected(obj)
    % errorWhenDisconnected Throws an error if the channel is disconnected.
    
        % Check the state.
        if ~obj.isConnected()
            % Throw an error to prevent operation by the calling method.
            error(message('vnt:XCP:ChannelDisconnected'));
        end
    end
    
    function errorWhenMeasurementActive(obj)
    % errorWhenMeasurementActive Throws an error if measurements are active.
    
        % Check the state.
        if obj.isMeasurementRunning()
            % Throw an error to prevent operation by the calling method.
            error(message('vnt:XCP:MeasurementActive'));
        end
    end
    
    function errorWhenMeasurementInactive(obj)
    % errorWhenMeasurementInactive Throws an error if measurements are inactive.
    
        % Check the state.
        if ~obj.isMeasurementRunning()
            % Throw an error to prevent operation by the calling method.
            error(message('vnt:XCP:MeasurementInactive'));
        end
    end
    
    function errorWhenNoMeasurementListsConfigured(obj)
    % errorWhenNoMeasurementListsConfigured Throws an error if no lists are configured.
        
        % Check if any DAQ or STIM lists are currently configured.
        if isempty(obj.DAQListObject) && isempty(obj.STIMListObject)
            % Throw an error to prevent operation by the calling method.
            error(message('vnt:XCP:NoMeasurementsListsConfigured'));
        end
    end
   
    function data = readFrame(obj, address, ECUAddressExtension, size)
    % readValue Read variable value.
        
        % Validate current state conditions.
        obj.errorWhenDisconnected();
        
        % Set the memory transfer address for this variable.
        obj.ProtocolObject.setMTA(ECUAddressExtension, address);
        data = [];
        
        % Upload data at MaxCTO - 1 at a time from server.
        for k = 1:floor(double(size) / double(obj.ServerInfo.MaxCTO-1))
            response = obj.ProtocolObject.upload(obj.ServerInfo.MaxCTO-1);
            % Exclude padding at end of response. g2248259.
            data = [data response(2:obj.ServerInfo.MaxCTO)]; %#ok<AGROW>
        end

        % Upload the last part (if any).
        rsize = rem(size, obj.ServerInfo.MaxCTO-1);
        if rsize > 0
            response = obj.ProtocolObject.upload(rsize);
            data = [data response(2:rsize+1)];
        end
    end
    
    function writeFrame(obj, frame, address, ECUAddressExtension, size)
    % writeValue Write variable value.
        
        % Validate current state conditions.
        obj.errorWhenDisconnected();
        
        % Set the memory transfer address for this variable.
        obj.ProtocolObject.setMTA(ECUAddressExtension, address);
        
        % Download data at MaxCTO-2 at a time to the server.
        frame_count = floor(double(size) / double(obj.ServerInfo.MaxCTO-2));
        for k = 1:frame_count
            start_idx = (k-1)*(obj.ServerInfo.MaxCTO-2)+1;
            end_idx = k*(obj.ServerInfo.MaxCTO-2);
            obj.ProtocolObject.download((obj.ServerInfo.MaxCTO-2), frame(start_idx:end_idx));
        end
        
        rsize = rem(size, (obj.ServerInfo.MaxCTO-2));
        if rsize > 0
            obj.ProtocolObject.download(rsize, frame(frame_count*(obj.ServerInfo.MaxCTO-2)+1:end));
        end
        
    end
    
end


methods % Custom get/set methods for properties.
    
    function value = get.NumberOfDAQLists(obj)
        value = numel(obj.DAQListObject);
    end
    
    function value = get.NumberOfSTIMLists(obj)
        value = numel(obj.STIMListObject);
    end
    
    function value = get.NumberOfMeasurementLists(obj)
        value = numel(obj.DAQListObject) + numel(obj.STIMListObject);
    end
    
    function set.SeedKeyDLL(obj, newValue)
        % Validate the new value argument.
        validateattributes(newValue, {'char'}, {'nonempty'});

        % Get the full path to the seed and key DLL file.
        [status, fileInfo, ~] = fileattrib(newValue);
        % Make sure the file exists.
        if status == 0
            error(message('vnt:XCP:SeedKeyDLLFileNotFound'));
        end
        
        % Set the property value.
        obj.SeedKeyDLL = newValue;
        % Also set a reference to the full path to the file.
        obj.SeedKeyDLLFullPath = fileInfo.Name; %#ok<MCSUP>
    end
    function  enableDAQProcessing(obj)
         obj.api_xcp_daqObject.enableDAQProcessing();
    end
     function disableDAQProcessing(obj)
	  obj.api_xcp_daqObject.disableDAQProcessing();
    end
    function out = getProtocolObject(obj)
        out = obj.ProtocolObject;
    end
    function out  = getApiXcpDAQObject(obj) 
        out  = obj.api_xcp_daqObject;
    end 
    function out = getDAQClockResp(obj)
        out = obj.DAQClockResp;
    end
    function out = getResolutionRespInfo(obj)
        out = obj.GetResolutionResp;
    end
end
    
end
