classdef Protocol < handle
% Protocol Class that implements the XCP protocol.

% Copyright 2012-2021 The MathWorks, Inc.

properties (SetAccess = 'private')
    A2LFileObject          % An A2L object.
    TransportLayerInfo     % A structure denoting transport layer information.
    TransportLayerObject   % An object reference to the transport layer implementation. For CAN, this is a can.Channel object.
    ReceptionTimer         % A timer object that runs the channel periodic tasks.
    ReceptionTimerListener % Listener for timer events.
    ReceptionTimerRate     % The rate at which to run the message reception task.
    State                  % The current processing state of the object.
    CurrentCommand         % A string denoting the current command sent.
    ResponseAvailable      % A boolean value denoting that command response data is available.
    ResponseData           % An array of bytes that are the most recently received command response data.
    ByteOrder              % The byte ordering of parameters in sent commands.
    ReceiverOverrun        % Flag indicating that the receiver callback has experienced overrun.
end

properties
    DAQData                % Storage for acquired DAQ data values.
end


methods
    
    function obj = Protocol(a2l, transportLayer, info)
    % Protocol Creates and returns an XCP Protocol object.    
        
        % Initialize properties.
        obj.ResponseAvailable = false;
        obj.DAQData = {};
        
        % Set the A2L file.
        obj.A2LFileObject = a2l;
        
        % Set the transport layer information.
        obj.TransportLayerInfo.Type = transportLayer;
        switch char(transportLayer)
            case 'CAN'
                obj.TransportLayerObject = xcp.transportlayer.CAN(info, obj.A2LFileObject.TransportLayerCANInfo);
                
            case 'CAN FD'
                obj.TransportLayerObject = xcp.transportlayer.CANFD(info, obj.A2LFileObject.TransportLayerCANInfo);
                
            case 'TCP'
                obj.TransportLayerObject = xcp.transportlayer.TCP(info.AddressString, info.Port);
                
            case 'UDP'
                obj.TransportLayerObject = xcp.transportlayer.UDP(info.AddressString, info.Port);
        end
        
        % Enable the transport layer.
        obj.TransportLayerObject.start();
            % Configure a timer for updating the message table. 
      %  obj.ReceptionTimerRate = 0.025;
      %  obj.ReceptionTimer = internal.IntervalTimer(obj.ReceptionTimerRate);
      %  obj.ReceptionTimerListener = event.listener( ...
      %      obj.ReceptionTimer, ...
       %     'Executing', ...
       %     @(~, ~)(obj.executeReceptionTimerCallback));
    
        
        % Set the ReceiverOverrun flag to false.
        obj.ReceiverOverrun = false;
        
        % Start the timer to begin processing.
      %  start(obj.ReceptionTimer);
    end

    function delete(obj)
        % Clear the timer if it exists.
        if ~isempty(obj.ReceptionTimer)
            % Stop the timer if running.
            if obj.ReceptionTimer.isRunning()
                stop(obj.ReceptionTimer);
            end            
            % Clear the timer property.
            obj.ReceptionTimer = [];        
        end
        
        % Clear the transport layer channel.
        % Check if the channel exists.
        if ~isempty(obj.TransportLayerObject)
            % Stop the channel if running.
            if obj.TransportLayerObject.Running
                stop(obj.TransportLayerObject);
            end
            % Delete the channel.
            delete(obj.TransportLayerObject);
        end
        % Clear the channel property.
        obj.TransportLayerObject = [];
    end

    
%% XCP Command Functions, Standard Commands, Mandatory

    function [response, resource, commModeBasic, maxCTO, maxDTO] = connect(obj, mode)
        % Set the current command property.
        obj.CurrentCommand = 'CONNECT';
        % Build the command data.
        cmdData = [0xFFu8 mode];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
        
        % Parse the resource information.
        resource.CAL = bitget(response(2), 1);
        resource.DAQ = bitget(response(2), 3);
        resource.PGM = bitget(response(2), 5);
        resource.STIM = bitget(response(2), 4);
        
        % Parse the communication mode information.
        addressGranularity = dec2bin(response(3), 8);
        addressGranularity = bin2dec(addressGranularity(6:7));
        switch addressGranularity
            case 0
                commModeBasic.AddressGranularity = 'BYTE';
            case 1
                commModeBasic.AddressGranularity = 'WORD';
            case 2
                commModeBasic.AddressGranularity = 'DWORD'; 
            case 3
                commModeBasic.AddressGranularity = 'RESERVED';
        end
        commModeBasic.BlockMode = bitget(response(3), 7);
        if bitget(response(3), 1)
            commModeBasic.ByteOrder = 'MSB_FIRST';
        else
            commModeBasic.ByteOrder = 'MSB_LAST';
        end
        
        % Set the byte order in this protocol object. This information is 
        % necessary for other commands sent through this object, so it must
        % be retained.
        obj.ByteOrder = commModeBasic.ByteOrder;

        % Parse the rest of the information.
        maxCTO = uint32(response(4));
        maxDTO = uint32(xcp.Utility.unpackParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_WORD, ...
            response(5:6)));
    end

    function response = disconnect(obj)
        % Set the current command property.
        obj.CurrentCommand = 'DISCONNECT';
        % Build the command data.
        cmdData = 0xFEu8;
        % Send the command.
        response = obj.transmitCommand(cmdData);
        
        % If empty is returned from the transmit attempt, then just return
        % from this function. According to XCP specification, there is
        % no positive response to this command.
        if isempty(response)
            return
        end
        
        % Only check for an error condition from disconnect if an actual
        % error code is received. 
        if response(1) == xcp.Utility.CODE_FOR_NEGATIVE_RESPONSE
            obj.validateResponse(obj.CurrentCommand, response);
        end
    end

    function [response, currentSessionStatus, currentResourceProtectionStatus, sessionConfigurationID] = getStatus(obj)
        % Set the current command property.
        obj.CurrentCommand = 'GET_STATUS';
        % Build the command data.
        cmdData = 0xFDu8;
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
        
        % Parse the current session status.
        currentSessionStatus.ClearDAQReq = bitget(response(2), 4);
        currentSessionStatus.DAQRunning = bitget(response(2), 7);
        currentSessionStatus.Resume = bitget(response(2), 8);
        currentSessionStatus.StoreCALReq = bitget(response(2), 1);
        currentSessionStatus.StoreDAQReq = bitget(response(2), 3);

        % Parse the current resource protection status.
        currentResourceProtectionStatus.CAL = bitget(response(3), 1);
        currentResourceProtectionStatus.DAQ = bitget(response(3), 3);
        currentResourceProtectionStatus.PGM = bitget(response(3), 5);
        currentResourceProtectionStatus.STIM = bitget(response(3), 4);        

        % Parse the rest of the information.
        sessionConfigurationID = xcp.Utility.unpackParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_WORD, ...
            response(5:6)); 
    end
    
    function response = synch(obj)
        % Set the current command property.
        obj.CurrentCommand = 'SYNCH';
        % Build the command data.
        cmdData = 0xFCu8;
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end

%% XCP Command Functions, Standard Commands, Optional

    function response = getCommModeInfo(obj)
        % Set the current command property.
        obj.CurrentCommand = 'GET_COMM_MODE_INFO';
        % Build the command data.
        cmdData = 0xFBu8;
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end

    function response = getID(obj, requestedIdentificationType)
        % Set the current command property.
        obj.CurrentCommand = 'GET_ID';
        % Build the command data.
        cmdData = [0xFAu8 requestedIdentificationType];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end

    function response = setRequest(obj, mode, sessionConfigurationID)
        % Set the current command property.
        obj.CurrentCommand = 'SET_REQUEST';
        % Build the command data.
        sessionConfigurationID = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_WORD, ...
            sessionConfigurationID);
        cmdData = [0xF9u8 mode sessionConfigurationID];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end

    function [response, seedLength, seedBytes] = getSeed(obj, modePart, modeResource)
        % Set the current command property.
        obj.CurrentCommand = 'GET_SEED';
        % Build the command data.
        cmdData = [0xF8u8 modePart modeResource];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
        
        % Parse the seed information.
        seedLength = response(2);
        seedBytes = response(3:end);
   end

    function [response, protectionStatus] = unlock(obj, length, key)
        % Set the current command property.
        obj.CurrentCommand = 'UNLOCK';
        % Build the command data.
        cmdData = [0xF7u8 length key];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
        
        % Parse the other information.
        protectionStatus = response(2);
    end

    function response = setMTA(obj, addressExtension, address)
        % Set the current command property.
        obj.CurrentCommand = 'SET_MTA';
        % Build the command data.
        address = obj.packAddressData(address);
        cmdData = [0xF6u8 0 0 addressExtension address];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end

    function response = upload(obj, numberOfDataElements)
        % Set the current command property.
        obj.CurrentCommand = 'UPLOAD';
        % Build the command data.
        cmdData = [0xF5u8 numberOfDataElements];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end
    
    function response = shortUpload(obj, numberOfDataElements, addressExtension, address)
        % Set the current command property.
        obj.CurrentCommand = 'SHORT_UPLOAD';
        % Build the command data.
        address = obj.packAddressData(address);
        cmdData = [0xF4u8 numberOfDataElements 0 addressExtension address];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end
    
    function response = buildChecksum(obj, blockSize)
        % Set the current command property.
        obj.CurrentCommand = 'BUILD_CHECKSUM';
        % Build the command data.
        blockSize = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_DWORD, ...
            blockSize);
        cmdData = [0xF3u8 blockSize];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end
    
%% XCP Command Functions, Auxiliary Commands

    function response = transportLayerCmd(obj, subCommandCode, parameters)
        % Set the current command property.
        obj.CurrentCommand = 'TRANSPORT_LAYER_CMD';
        % Build the command data.
        cmdData = [0xF2u8 subCommandCode parameters];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end

    function response = userCmd(obj, subCommandCode, parameters)
        % Set the current command property.
        obj.CurrentCommand = 'USER_CMD';
        % Build the command data.
        cmdData = [0xF1u8 subCommandCode parameters];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end

%% XCP Command Functions, DAQ/STIM Commands, Mandatory
    
    function response = setDAQPtr(obj, DAQListNumber, ODTNumber, ODTEntryNumber)
        % Set the current command property.
        obj.CurrentCommand = 'SET_DAQ_PTR';
        % Build the command data.
        DAQListNumber = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_WORD, ...
            DAQListNumber);
        cmdData = [0xE2u8 0 DAQListNumber ODTNumber ODTEntryNumber];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end
    
    function response = writeDAQ(obj, bitOffset, DAQElementSize, addressExtension, address)
        % Set the current command property.
        obj.CurrentCommand = 'WRITE_DAQ';
        % Build the command data.
        address = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_DWORD, ...
            address);
        cmdData = [0xE1u8 bitOffset DAQElementSize addressExtension address];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end
    
    function response = setDAQListMode(obj, mode, DAQListNumber, eventChannelNumber, ratePrescaler, DAQListPriority)
        % Set the current command property.
        obj.CurrentCommand = 'SET_DAQ_LIST_MODE';
        % Build the command data.
        DAQListNumber = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_WORD, ...
            DAQListNumber);
        eventChannelNumber = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_WORD, ...
            eventChannelNumber);        
        cmdData = [0xE0u8 mode DAQListNumber eventChannelNumber ratePrescaler DAQListPriority];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end
    
    function [response, firstPID] = startStopDAQList(obj, mode, DAQListNumber)
        % Set the current command property.
        obj.CurrentCommand = 'START_STOP_DAQ_LIST';
        % Build the command data.
        DAQListNumber = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_WORD, ...
            DAQListNumber);  
        cmdData = [0xDEu8 mode DAQListNumber];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
        
        % Parse the first PID value.
        firstPID = response(2);
    end
    
    function response = startStopSynch(obj, mode)
         % Set the current command property.
        obj.CurrentCommand = 'START_STOP_SYNCH';
       % Build the command data.
        cmdData = [0xDDu8 mode];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end

%% XCP Command Functions, DAQ/STIM Commands, Optional

    function response = readDAQ(obj)
        % Set the current command property.
        obj.CurrentCommand = 'READ_DAQ';
        % Build the command data.
        cmdData = 0xDBu8;
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end

    function response = getDAQClock(obj)
        % Set the current command property.
        obj.CurrentCommand = 'GET_DAQ_CLOCK';
        % Build the command data.
        cmdData = 0xDCu8;
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end
    
    function response = getDAQProcessorInfo(obj)
        % Set the current command property.
        obj.CurrentCommand = 'GET_DAQ_PROCESSOR_INFO';
        % Build the command data.
        cmdData = 0xDAu8;
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end
    
    function response = getDAQResolutionInfo(obj)
        % Set the current command property.
        obj.CurrentCommand = 'GET_DAQ_RESOLUTION_INFO';
        % Build the command data.
        cmdData = 0xD9u8;
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end
    
    function response = getDAQListMode(obj, DAQListNumber)
        % Set the current command property.
        obj.CurrentCommand = 'GET_DAQ_LIST_MODE';
        % Build the command data.
        DAQListNumber = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_WORD, ...
            DAQListNumber);  
        cmdData = [0xDFu8 0 DAQListNumber];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end
    
    function response = getDAQEventInfo(obj, eventChannelNumber)
        % Set the current command property.
        obj.CurrentCommand = 'GET_DAQ_EVENT_INFO';
        % Build the command data.
        eventChannelNumber = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_WORD, ...
            eventChannelNumber); 
        cmdData = [0xD7u8 0 eventChannelNumber];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end
    
%% XCP Command Functions, Dynamic DAQ List Configuration Commands, Mandatory
    
    function response = freeDAQ(obj)
        % Set the current command property.
        obj.CurrentCommand = 'FREE_DAQ';
        % Build the command data.
        cmdData = 0xD6u8;
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end

    function response = allocDAQ(obj, DAQCount)
        % Set the current command property.
        obj.CurrentCommand = 'ALLOC_DAQ';
        % Build the command data.
        DAQCount = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_WORD, ...
            DAQCount); 
        cmdData = [0xD5u8 0 DAQCount];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end
    
    function response = allocODT(obj, DAQListNumber, ODTCount)
        % Set the current command property.
        obj.CurrentCommand = 'ALLOC_ODT';
        % Build the command data.
        DAQListNumber = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_WORD, ...
            DAQListNumber); 
        cmdData = [0xD4u8 0 DAQListNumber ODTCount];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end

    function response = allocODTEntry(obj, DAQListNumber, ODTNumber, ODTEntriesCount)
        % Set the current command property.
        obj.CurrentCommand = 'ALLOC_ODT_ENTRY';
        % Build the command data.
        DAQListNumber = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_WORD, ...
            DAQListNumber); 
        cmdData = [0xD3u8 0 DAQListNumber ODTNumber ODTEntriesCount];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end

%% XCP Command Functions, Dynamic DAQ List Configuration Commands, Mandatory
    function response = download(obj, numberOfDataElements, data)
        % Set the current command property.
        obj.CurrentCommand = 'DOWNLOAD';
        % Build the command data.
        cmdData = [0xF0u8 numberOfDataElements data];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end

%% XCP Command Functions, Static DAQ List Configuration Commands, Mandatory
    function response = clearDAQList(obj, DAQListNumber)
        % Set the current command property.
        obj.CurrentCommand = 'CLEAR_DAQ_LIST';
        % Build the command data.
        DAQListNumber = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_WORD, ...
            DAQListNumber); 
        cmdData = [0xE2u8 0 DAQListNumber];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end

%% XCP Command Functions, Static DAQ List Configuration Commands, Optional
    function response = getDAQListInfo(obj, DAQListNumber)
        % Set the current command property.
        obj.CurrentCommand = 'GET_DAQ_LIST_INFO';
        % Build the command data.
        DAQListNumber = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_WORD, ...
            DAQListNumber); 
        cmdData = [0xD8u8 0 DAQListNumber];
        % Send the command.
        response = obj.transmitCommand(cmdData);
        % Check the response for an error condition.
        obj.validateResponse(obj.CurrentCommand, response);
    end
    
end % Public methods


methods (Access = 'private')

    function address = packAddressData( obj, address )
        
        address = xcp.Utility.packParameter( ...
            obj.ByteOrder, ...
            xcp.Utility.BYTES_IN_DWORD, ...
            address);
    end
    
    function validateResponse(obj, command, response) %#ok<INUSL>
        % Check for an empty response.
        if isempty(response)
            error(message('vnt:XCP:NoServerResponseReceived', command));
        end
        
        % If the reponse was anything other than an error response, return.
        if response(1) ~= xcp.Utility.CODE_FOR_NEGATIVE_RESPONSE
            return;
        end
        
        % For an error code, throw an error indicating the error code,
        % type, and detailed information about the error.
        error(message('vnt:XCP:ServerResponseError', ...
            command, ...
            xcp.Utility.MAP_ERROR_CODE_TO_NAME(response(2)), ...
            xcp.Utility.MAP_ERROR_CODE_TO_EXPLANATION(response(2))));
    end
    
    function response = transmitCommand(obj, cmdData, varargin)
    % transmitCommand Sends an XCP command message.
    % 
    %   cmdData - An array of bytes that are the command message code and
    %       data.
    %   varargin - An optional value indicating how many times the
    %       current command has been issued. On first transmit, this
    %       argument does not exist. It gets incremented for each
    %       subsequent transmit attempt.
        
        % Set initial return value as empty. This defaults the return value
        % to indicate no response was received for the command.
        response = [];
        
        % Send the command according to the configured transport layer.
        obj.TransportLayerObject.transmit(cmdData);
        obj.ResponseData= obj.TransportLayerObject.receive(inf);
           % is received. We hard code this to one second as A2L file
        % information is not always reliable.
       % tStart = tic;
       % while toc(tStart) <= 1
      %      % Check for a response and break the loop if found.
      %      if obj.ResponseAvailable
       %         break;
       %     end
       %     % Pause for some time.
       %     pause(0.010);
       % end

        % Check for a response or a timeout.
       
        if ~isempty(obj.ResponseData)
            % Set the response data from the shared property and clear it.
            response = obj.ResponseData;
            obj.ResponseData = [];
            % Reset the response available property.
            obj.ResponseAvailable = false;
        
        elseif nargin == 2
            % Some commands require special handling.
            switch obj.CurrentCommand
                case 'DISCONNECT'
                    % On a disconnect command, return immediately.
                    return;
                otherwise
                    % Retry the command for a second time via recursive call.
                    response = obj.transmitCommand(cmdData, 2);
            end
            
        elseif nargin == 3 && varargin{1} == 2
            % Retry the command for a third time via recursive call.
            response = obj.transmitCommand(cmdData, 3);
        end
    end
    
    function executeReceptionTimerCallback(obj)
    % executeReceptionTimerCallback Timer callback function for processing received messages.
        
        % Return immediately if there is no transport layer object to use.
        if isempty(obj.TransportLayerObject)
            return;
        end
        
        % Return immediately if this Protocol object has experienced overrun.
        if obj.ReceiverOverrun
            return;
        end
        
        % Get all of the available messages from the transport layer.
        msgRx = obj.TransportLayerObject.receive(inf);
        
        % If overrun has occurred in transport layer, set the ReceiverOverrun
        % flag to true on this Protocol object and return immediately.
        if obj.TransportLayerObject.OverrunFlag
            obj.ReceiverOverrun = true;
            return;
        end
        
        % Return immediately if no messages were received.
        if isempty(msgRx)
            return;
        end
        
        % Process the received messages by analyzing the PID.
        for ii = 1:numel(msgRx)
            switch msgRx(ii).Data(1)
                case {xcp.Utility.CODE_FOR_POSITIVE_RESPONSE,xcp.Utility.CODE_FOR_NEGATIVE_RESPONSE}
                    % On a positive or negative response, set the response 
                    % indication properties.
                    obj.ResponseAvailable = true;
                    % Process all valid received messages.
                    % Fix the bug that only the latest received message was
                    % processed as the response data. g2554994.
                    if isempty(obj.ResponseData)
                        obj.ResponseData = msgRx(ii).Data;
                    else
                        obj.ResponseData = [obj.ResponseData, msgRx(ii).Data(2:end)];
                    end
                case xcp.Utility.CODE_FOR_SERVER_EVENT
                case xcp.Utility.CODE_FOR_SERVER_SERVICE_REQUEST
                otherwise
                    % On a DAQ DTO, log the message for further processing.
                    obj.DAQData{end + 1} = msgRx(ii).Data;
            end
        end
    end
    
end % Private methods
    
end % Class
