function obj = xcpChannel(varargin)
% xcpChannel Creates an XCP channel able to connect to a server module.
%
%   OBJ = xcpChannel(A2L_OBJECT, TRANSPORT_LAYER, DEVICE_VENDOR, DEVICE, DEVICE_CHANNEL_INDEX)
%   OBJ = xcpChannel(A2L_OBJECT, TRANSPORT_LAYER, DEVICE_VENDOR, DEVICE)
%   creates an XCP channel OBJ connected to the server device described
%   by the A2L file A2L_OBJECT. In this form, TRANSPORT_LAYER is specified
%   as 'CAN' or 'CAN FD'. The device selection arguments when using CAN or
%   CAN FD depend on the vendor of the device. DEVICE_VENDOR and DEVICE
%   must always be provided. DEVICE_CHANNEL_INDEX is required for some
%   vendors.
%
%   OBJ = xcpChannel(A2L_OBJECT, TRANSPORT_LAYER) creates an XCP channel OBJ
%   connected to the server device described by the A2L file A2L_OBJECT.
%   In this form, TRANSPORT_LAYER is specified as 'UDP' or 'TCP' to use
%   ethernet connections communicate with the server. The server IP address
%   and port values are taken from the A2L file.
%
%   OBJ = xcpChannel(A2L_OBJECT, TRANSPORT_LAYER, IP_ADDRESS, PORT) creates
%   an XCP channel OBJ connected to the server device described by the
%   A2L file A2L_OBJECT. In this form, TRANSPORT_LAYER is specified as
%   'UDP' or 'TCP' to use ethernet connections communicate with the
%   server. The server IP_ADDRESS and PORT values are provided as input.
%
%   Example:
%       a2lObj = xcpA2L('myFile.a2l');
%
%       channelObj = xcpChannel(a2lObj, 'CAN', 'Vector', 'VN1610 1', 1);
%       channelObj = xcpChannel(a2lObj, 'CAN FD', 'Vector', 'VN1610 1', 1);
%
%       channelObj = xcpChannel(a2lObj, 'CAN', 'NI', 'CAN1');
%       channelObj = xcpChannel(a2lObj, 'CAN FD', 'NI', 'CAN1');
%
%       channelObj = xcpChannel(a2lObj, 'UDP');
%       channelObj = xcpChannel(a2lObj, 'TCP');
%
%       channelObj = xcpChannel(a2lObj, 'UDP', '192.168.1.50', 5555);
%       channelObj = xcpChannel(a2lObj, 'TCP', '192.168.1.50', 5555);
%
%   See also VNT.

% Copyright 2013-2021 The MathWorks, Inc.

% Perform a Windows OS check.
%if ~ispc()
%    error(message('vnt:XCP:WindowsOnly'));
%end

% Perform an argument count check.
narginchk(2, 5);

% Convert string inputs to character vectors.
[varargin{:}] = convertStringsToChars(varargin{:});

% Pass the call through to the xcp.Channel class.
obj = xcp.Channel(varargin{:});
end
